﻿using SimpleJSON;
using System;
using System.Net;
using System.Net.Sockets;
using System.Text;
using ToySerialController.UI;
using ToySerialController.Utils;
using UnityEngine;

namespace ToySerialController.Device.OutputTarget
{
    public class UdpOutputTarget : IOutputTarget
    {
        private UITextInput PortInput;
        private JSONStorableString PortText;
        private UIHorizontalGroup ButtonGroup;

        private UdpClient _client;

        public void CreateUI(IUIBuilder builder)
        {
            PortInput = builder.CreateTextInput("OutputTarget:Udp:Port", "Port:", "8889", 50);
            PortText = PortInput.storable;

            ButtonGroup = builder.CreateHorizontalGroup(510, 50, new Vector2(10, 0), 2, idx => builder.CreateButtonEx());
            var startSerialButton = ButtonGroup.items[0].GetComponent<UIDynamicButton>();
            startSerialButton.label = "Start Udp";
            startSerialButton.button.onClick.AddListener(StartUdp);
            UIBuilder.CreateAction("StartUdp", StartUdp);

            var stopSerialButton = ButtonGroup.items[1].GetComponent<UIDynamicButton>();
            stopSerialButton.label = "Stop Udp";
            stopSerialButton.button.onClick.AddListener(StopUdp);
            UIBuilder.CreateAction("StopUdp", StopUdp);
        }

        public void DestroyUI(IUIBuilder builder)
        {
            builder.Destroy(PortInput);
            builder.Destroy(ButtonGroup);
        }

        public void RestoreConfig(JSONNode config)
        {
            config.Restore(PortText);
        }

        public void StoreConfig(JSONNode config)
        {
            config.Store(PortText);
        }

        private void StartUdp()
        {
            if (_client != null)
                return;

            try
            {
                var port = int.Parse(PortText.val);
                var endpoint = new IPEndPoint(IPAddress.Loopback, port);
                _client = new UdpClient
                {
                    ExclusiveAddressUse = false
                };

                _client.Client.Blocking = false;
                _client.Client.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReuseAddress, true);
                _client.Client.Bind(endpoint);
                _client.Connect(endpoint);

                SuperController.LogMessage($"Upd started on port: {port}");
            }
            catch (Exception e)
            {
                SuperController.LogError(e.ToString());
            }
        }

        private void StopUdp()
        {
            if (_client == null)
                return;

            try
            {
                _client.Close();
            }
            catch(Exception e)
            {
                SuperController.LogError(e.ToString());
            }

            _client = null;

            SuperController.LogMessage("Upd stopped");
        }


        public void Write(string data)
        {
            if (_client == null)
                return;

            var bytes = Encoding.ASCII.GetBytes(data);
            var sent = _client.Send(bytes, bytes.Length);
        }
        public void Update(bool tracking)
        {

        }

        public void OnSceneChanging() { }
        public void OnSceneChanged() { }

        protected virtual void Dispose(bool disposing)
        {
            StopUdp();
        }

        public void Dispose()
        {
            Dispose(disposing: true);
            GC.SuppressFinalize(this);
        }
    }
}
