﻿using SimpleJSON;
using System.Collections.Generic;
using System.Linq;
using ToySerialController;
using ToySerialController.UI;
using ToySerialController.Utils;
using UnityEngine;
using DebugUtils;
using System;
using Leap.Unity.Infix;
using Leap.Unity;

namespace ToySerialController.MotionSource
{
    public abstract class AbstractFemaleMotionSource : AbstractRefreshableMotionSource
    {
        private Atom _femaleAtom;
        private Vector3 _targetPosition;
        private Vector3 _targetUp;
        private Vector3 _targetRight;
        private Vector3 _targetForward;
        // private Atom _currentAutoFemaleAtom;

        private float _referenceLengthMultiplier = 1f;
        private float _referenceRadiusMultiplier = 1f;

        protected string _invalidAutoTargetUid = "";

        private JSONStorableStringChooser FemaleChooser;
        private JSONStorableStringChooser TargetChooser;
        private JSONStorableBool IncludeMales;

        private JSONStorableBool AutoTargetEnableVaginal;
        private JSONStorableBool AutoTargetEnableMouth;
        private JSONStorableBool AutoTargetEnableAnus;
        private JSONStorableBool AutoTargetEnableLeftHand;
        private JSONStorableBool AutoTargetEnableRightHand;
        private JSONStorableBool AutoTargetEnableChest;
        private JSONStorableBool AutoTargetEnableLeftFoot;
        private JSONStorableBool AutoTargetEnableRightFoot;
        private JSONStorableBool AutoTargetEnableFeet;
        private JSONStorableBool AutoTargetPrioritizeVaginalOverAnus;
        private JSONStorableBool AutoTargetPrioritizeAnusOverVaginal;

        private SuperController Controller => SuperController.singleton;

        public override Vector3 TargetPosition => _targetPosition;
        public override Vector3 TargetUp => _targetUp;
        public override Vector3 TargetRight => _targetRight;
        public override Vector3 TargetForward => _targetForward;

        public override float ReferenceLengthMultiplier { get { return _referenceLengthMultiplier; } }
        public override float ReferenceRadiusMultiplier { get { return _referenceRadiusMultiplier; } }

        public override void CreateUI(IUIBuilder builder)
        {
            var targets = new List<string> { "Auto", "Vagina", "Pelvis", "Hips", "Anus", "Mouth", "Left Hand", "Right Hand", "Chest", "ChestExtended", "Left Foot", "Right Foot", "Feet" };
            var defaultTarget = "Auto";

            FemaleChooser = builder.CreatePopup("MotionSource:Female", "Select Female", null, null, FemaleChooserCallback);
            IncludeMales = builder.CreateToggle("MotionSource:IncludeMales", "Include Males In Female Targets", false, (b) => { RefreshButtonCallback(); });

            TargetChooser = builder.CreateScrollablePopup("MotionSource:FemaleTarget", "Select Target Point", targets, defaultTarget, null);

            AutoTargetEnableVaginal = builder.CreateToggle("MotionSource:AutoTargetEnableVaginal", "Auto: Enable Vaginal", true, (b) => { });
            AutoTargetPrioritizeVaginalOverAnus = builder.CreateToggle("MotionSource:AutoTargetPrioritizeVaginalOverAnus", "Auto: Prioritize Vaginal Over Anus", true, (b) => {
                if (b)
                {
                    AutoTargetPrioritizeAnusOverVaginal.val = false;
                }
            });
            AutoTargetEnableAnus = builder.CreateToggle("MotionSource:AutoTargetEnableAnus", "Auto: Enable Anus", true, (b) => { });
            AutoTargetPrioritizeAnusOverVaginal = builder.CreateToggle("MotionSource:AutoTargetPrioritizeAnusOverVaginal", "Auto: Prioritize Anus Over Vaginal", false, (b) => { 
                if (b)
                {
                    AutoTargetPrioritizeVaginalOverAnus.val = false;
                }
            });
            AutoTargetEnableMouth = builder.CreateToggle("MotionSource:AutoTargetEnableMouth", "Auto: Enable Mouth", true, (b) => { });
            AutoTargetEnableLeftHand = builder.CreateToggle("MotionSource:AutoTargetEnableLeftHand", "Auto: Enable Left Hand", true, (b) => { });
            AutoTargetEnableRightHand = builder.CreateToggle("MotionSource:AutoTargetEnableRightHand", "Auto: Enable Right Hand", true, (b) => { });
            AutoTargetEnableChest = builder.CreateToggle("MotionSource:AutoTargetEnableChest", "Auto: Enable Chest", true, (b) => { });
            AutoTargetEnableLeftFoot = builder.CreateToggle("MotionSource:AutoTargetEnableLeftFoot", "Auto: Enable Left Foot", true, (b) => { });
            AutoTargetEnableRightFoot = builder.CreateToggle("MotionSource:AutoTargetEnableRightFoot", "Auto: Enable Right Foot", true, (b) => { });
            AutoTargetEnableFeet = builder.CreateToggle("MotionSource:AutoTargetEnableFeet", "Auto: Enable Feet", true, (b) => { });

            FindFemales();

            base.CreateUI(builder);
        }

        public override void DestroyUI(IUIBuilder builder)
        {
            builder.Destroy(FemaleChooser);
            builder.Destroy(TargetChooser);
            builder.Destroy(IncludeMales);
            builder.Destroy(AutoTargetEnableVaginal);
            builder.Destroy(AutoTargetEnableMouth);
            builder.Destroy(AutoTargetEnableAnus);
            builder.Destroy(AutoTargetEnableLeftHand);
            builder.Destroy(AutoTargetEnableRightHand);
            builder.Destroy(AutoTargetEnableChest);
            builder.Destroy(AutoTargetEnableLeftFoot);
            builder.Destroy(AutoTargetEnableRightFoot);
            builder.Destroy(AutoTargetEnableFeet);
            builder.Destroy(AutoTargetPrioritizeVaginalOverAnus);
            builder.Destroy(AutoTargetPrioritizeAnusOverVaginal);

            base.DestroyUI(builder);
        }

        public override void StoreConfig(JSONNode config)
        {
            config.Store(IncludeMales);
            config.Store(FemaleChooser);
            config.Store(TargetChooser);
            config.Store(AutoTargetEnableVaginal);
            config.Store(AutoTargetEnableMouth);
            config.Store(AutoTargetEnableAnus);
            config.Store(AutoTargetEnableLeftHand);
            config.Store(AutoTargetEnableRightHand);
            config.Store(AutoTargetEnableChest);
            config.Store(AutoTargetEnableLeftFoot);
            config.Store(AutoTargetEnableRightFoot);
            config.Store(AutoTargetEnableFeet);
            config.Store(AutoTargetPrioritizeVaginalOverAnus);
            config.Store(AutoTargetPrioritizeAnusOverVaginal);
        }

        public override void RestoreConfig(JSONNode config)
        {
            config.Restore(IncludeMales);
            config.Restore(FemaleChooser);
            config.Restore(TargetChooser);
            config.Restore(AutoTargetEnableVaginal);
            config.Restore(AutoTargetEnableMouth);
            config.Restore(AutoTargetEnableAnus);
            config.Restore(AutoTargetEnableLeftHand);
            config.Restore(AutoTargetEnableRightHand);
            config.Restore(AutoTargetEnableChest);
            config.Restore(AutoTargetEnableLeftFoot);
            config.Restore(AutoTargetEnableRightFoot);
            config.Restore(AutoTargetEnableFeet);
            config.Restore(AutoTargetPrioritizeVaginalOverAnus);
            config.Restore(AutoTargetPrioritizeAnusOverVaginal);

            FindFemales();
        }

        private string _lastAutoBestFemaleUid = "";
        public override bool Update(Func<IMotionSource, float> motionTestFunc)
        {
            if (FemaleChooser.val == FEMALE_AUTO)
            {
                // _currentAutoFemaleAtom = _femaleAtom;

                Atom bestFemale = null;
                Vector3 bestTargetPosition = _targetPosition;
                Vector3 bestTargetUp = _targetUp;
                Vector3 bestTargetRight = _targetRight;
                Vector3 bestTargetForward = _targetForward;
                float bestReferenceLengthMultiplier = _referenceLengthMultiplier;
                float bestReferenceRadiusMultiplier = _referenceRadiusMultiplier;
                float bestMotionScore = float.MinValue;
                
                int firstIndexTest = 0;
                if (_femaleAtom != null)
                {
                    firstIndexTest = _femaleUids.IndexOf(_femaleAtom.uid);
                    if (firstIndexTest == -1)
                        firstIndexTest = 0;
                }

                for (int i = 0; i < _femaleUids.Count; i++)
                {
                    if (i == 0)
                        _femaleAtom = Controller.GetAtomByUid(_femaleUids[firstIndexTest]);
                    else if (i == firstIndexTest)
                        _femaleAtom = Controller.GetAtomByUid(_femaleUids[0]);
                    else
                        _femaleAtom = Controller.GetAtomByUid(_femaleUids[i]);

                    float motionScore = -1f;
                    if (UpdateCurrentFemale(motionTestFunc, out motionScore) && bestMotionScore < motionScore)
                    {
                        bestMotionScore = motionScore;
                        bestTargetPosition = _targetPosition;
                        bestTargetUp = _targetUp;
                        bestTargetRight = _targetRight;
                        bestTargetForward = _targetForward;
                        bestFemale = _femaleAtom;
                        bestReferenceLengthMultiplier = _referenceLengthMultiplier;
                        bestReferenceRadiusMultiplier = _referenceRadiusMultiplier;
                    }
                }

                // _currentAutoFemaleAtom = bestFemale;

                if (bestFemale != null)
                {
                    _femaleAtom = bestFemale;
                    _targetPosition = bestTargetPosition;
                    _targetUp = bestTargetUp;
                    _targetRight = bestTargetRight;
                    _targetForward = bestTargetForward;
                    _referenceLengthMultiplier = bestReferenceLengthMultiplier;
                    _referenceRadiusMultiplier = bestReferenceRadiusMultiplier;
                    if (_lastAutoBestFemaleUid != _femaleAtom.uid)
                    {
                        TCodeDevice.TargetSwappedTime = Time.time;
                        //if (ToySerialController.Plugin.DebugEnabled)
                        //    SuperController.LogMessage("TCodeDevice.TargetSwappedTime reset");

                        _lastAutoBestFemaleUid = _femaleAtom.uid;
                        if (ToySerialController.Plugin.DebugEnabled)
                            SuperController.LogMessage("Current Auto Person: " + _femaleAtom.uid);
                    }
                    return true;
                }
                _referenceLengthMultiplier = 1f;
                _referenceRadiusMultiplier = 1f;
            }
            else
            {
                float motionScore = -1f;
                return UpdateCurrentFemale(motionTestFunc, out motionScore);
            }

            return false;
        }

        private bool UpdateCurrentFemale(Func<IMotionSource, float> motionTestFunc, out float motionScore)
        {
            if (_femaleAtom == null || _femaleAtom.uid == _invalidAutoTargetUid || !_femaleAtom.on)
            {
                motionScore = float.MinValue;
                return false;
            }

            if (UpdateTarget(motionTestFunc, out motionScore))
            {
                DebugDraw.DrawTransform(TargetPosition, TargetUp, TargetRight, TargetForward, 0.15f);
                return true;
            }

            return false;
        }

        private bool UpdateTarget(Func<IMotionSource, float> motionTestFunc, out float motionScore)
        {
            bool result = false;

            if (TargetChooser.val == "Auto") return UpdateAutoTarget(motionTestFunc, out motionScore);
            else if (TargetChooser.val == "Vagina") result = UpdateVaginaTarget();
            else if (TargetChooser.val == "Pelvis") result = UpdateFreeControllerTarget("pelvisControl");
            else if (TargetChooser.val == "Hips") result = UpdateFreeControllerTarget("hipControl");
            else if (TargetChooser.val == "Anus") result = UpdateAnusTarget();
            else if (TargetChooser.val == "Mouth") result = UpdateMouthTarget();
            else if (TargetChooser.val == "Left Hand") result = UpdateLeftHandTarget();
            else if (TargetChooser.val == "Right Hand") result = UpdateRightHandTarget();
            else if (TargetChooser.val == "Chest") result = UpdateChestTarget();
            else if (TargetChooser.val == "ChestExtended") result = UpdateChestExtendedTarget();
            else if (TargetChooser.val == "Left Foot") result = UpdateLeftFootTarget();
            else if (TargetChooser.val == "Right Foot") result = UpdateRightFootTarget();
            else if (TargetChooser.val == "Feet") result = UpdateFeetTarget();

            motionScore = 0f; // Other targets than "Auto" will always return 0 in score.

            return result;
        }

        private class PersonAutoTargetData
        {
            public List<AutoTargetData> AutoTargets;
            public int LastBestIndex = -1;
            public int LastBestIndexConsistencyScore = 0;

            public PersonAutoTargetData(List<AutoTargetData> autoTargets)
            {
                AutoTargets = autoTargets;
            }
        }

        private class AutoTargetData
        {
            public string TargetInfo;
            public Func<bool> TargetFunc;
            
            public Vector3 LastPosition;
            public float[] RecordedPositionsMagnitudes;
            public int RecordedIndex = 0;
            public const int SAMPLING_COUNT = 30;
            public bool Enabled;

            public AutoTargetData(Func<bool> func, string info)
            {
                TargetFunc = func;
                TargetInfo = info;
                Enabled = true;

                RecordedPositionsMagnitudes = new float[SAMPLING_COUNT];
                for (int i = 0; i < SAMPLING_COUNT; i++)
                    RecordedPositionsMagnitudes[i] = 0f;
                RecordedIndex = -1;
                LastPosition = Vector3.zero;
            }

            public void RecordPosition(Vector3 pos)
            {
                float m = (LastPosition - pos).magnitude * 100f;
                if (RecordedIndex == -1)
                {
                    m = 0f; // first record
                    RecordedIndex = 0;
                }
                RecordedPositionsMagnitudes[RecordedIndex] = m;
                RecordedIndex = (RecordedIndex + 1) % SAMPLING_COUNT;
                LastPosition = pos;
            }

            public float GetAverageMagnitude()
            {
                return (RecordedPositionsMagnitudes.Sum() / (float)SAMPLING_COUNT);
            }
        }

        private Dictionary<string, PersonAutoTargetData> _targetsPerPerson;
        private float _lastTrackedVaginal = 0f;
        private float _lastTrackedAnus = 0f;

        private bool UpdateAutoTarget(Func<IMotionSource, float> motionTestFunc, out float motionScore)
        {
            if (_targetsPerPerson == null)
                _targetsPerPerson = new Dictionary<string, PersonAutoTargetData>();

            PersonAutoTargetData personAutoTargetData = null;
            if (!_targetsPerPerson.TryGetValue(_femaleAtom.uid, out personAutoTargetData))
            {
                personAutoTargetData = new PersonAutoTargetData(new List<AutoTargetData>
                {
                    new AutoTargetData(UpdateVaginaTarget, nameof(UpdateVaginaTarget)),
                    new AutoTargetData(UpdateAnusTarget, nameof(UpdateAnusTarget)),
                    new AutoTargetData(UpdateMouthTarget, nameof(UpdateMouthTarget)),
                    new AutoTargetData(UpdateLeftHandTarget, nameof(UpdateLeftHandTarget)),
                    new AutoTargetData(UpdateRightHandTarget, nameof(UpdateRightHandTarget)),
                    new AutoTargetData(UpdateChestExtendedTarget, nameof(UpdateChestExtendedTarget)),
                    new AutoTargetData(UpdateFeetTarget, nameof(UpdateFeetTarget)),
                    new AutoTargetData(UpdateLeftFootTarget, nameof(UpdateLeftFootTarget)),
                    new AutoTargetData(UpdateRightFootTarget, nameof(UpdateRightFootTarget))
                });
                _targetsPerPerson.Add(_femaleAtom.uid, personAutoTargetData);
            }
            personAutoTargetData.AutoTargets[0].Enabled = AutoTargetEnableVaginal.val;
            personAutoTargetData.AutoTargets[1].Enabled = AutoTargetEnableAnus.val;
            personAutoTargetData.AutoTargets[2].Enabled = AutoTargetEnableMouth.val;
            personAutoTargetData.AutoTargets[3].Enabled = AutoTargetEnableLeftHand.val;
            personAutoTargetData.AutoTargets[4].Enabled = AutoTargetEnableRightHand.val;
            personAutoTargetData.AutoTargets[5].Enabled = AutoTargetEnableChest.val;
            personAutoTargetData.AutoTargets[6].Enabled = AutoTargetEnableFeet.val;
            personAutoTargetData.AutoTargets[7].Enabled = AutoTargetEnableLeftFoot.val;
            personAutoTargetData.AutoTargets[8].Enabled = AutoTargetEnableRightFoot.val;

            motionScore = float.MinValue;

            AutoTargetData resultingAutoTargetData = null;
            float bestDistance = float.MaxValue;
            int bestIndex = -1;

            AutoTargetData bestMotionAutoTargetData = null;
            float bestMotionScore = 0f;
            int bestMotionIndex = 0;

            bool consistencyScoreUpdated = false;

            bool vaginalOverAnusCurrentlyPrioritized = personAutoTargetData.LastBestIndex == 0 && AutoTargetPrioritizeVaginalOverAnus.val && (Time.time - _lastTrackedVaginal) < 1f;
            bool anusOverVaginalCurrentlyPrioritized = !vaginalOverAnusCurrentlyPrioritized && personAutoTargetData.LastBestIndex == 1 && AutoTargetPrioritizeAnusOverVaginal.val && (Time.time - _lastTrackedAnus) < 1f;

            for (int i = 0; i < personAutoTargetData.AutoTargets.Count; i ++)
            {
                AutoTargetData targetData = personAutoTargetData.AutoTargets[i];
                if (targetData.Enabled && targetData.TargetFunc())
                {
                    float distance = Vector3.Distance(ReferencePosition, TargetPosition);

                    // filter away non-relevant targets based on distance (optimization)
                    if (distance > 1f)
                    {
                        //if (ToySerialController.Plugin.DebugEnabled)
                        //    SuperController.LogMessage("Ignoring " + targetData.TargetInfo + " WAY TOO FAR away: " + distance.ToString());
                        continue;
                    }

                    // record local position difference in the motion between target and reference
                    Vector3 targetPosAlongReference = NearestPointOnLine(ReferencePosition, ReferenceUp, _targetPosition);
                    targetData.RecordPosition(targetPosAlongReference - ReferencePosition);

                    if (distance < bestDistance)
                    {
                        resultingAutoTargetData = targetData;
                        bestDistance = distance;
                        bestIndex = i;
                    }

                    if (motionTestFunc != null)
                    {
                        // Test function to get a score based on position. -1 = no hit. This motion test returns -1 or 1 atm.
                        float score = motionTestFunc(this);
                        if (score >= 0f)
                        {
                            // Add average motion magnitude to the score together with the alignment as a final score. 
                            // The better accuracy and movement => the higher priority the target will be
                            float d = Vector3.Dot(TargetUp, ReferenceUp);
                            float avm = targetData.GetAverageMagnitude();
                            score = score * d * Math.Max(0.05f, avm);

                            if (i == 0)
                            {
                                _lastTrackedVaginal = Time.time;

                                if (anusOverVaginalCurrentlyPrioritized)
                                {
                                    // skip vaginal as long as anus is active
                                    continue;
                                }
                            }
                            else if (i == 1)
                            {
                                _lastTrackedAnus = Time.time;
                                
                                if (vaginalOverAnusCurrentlyPrioritized)
                                {
                                    // skip anus as long as vaginal is active
                                    continue;
                                }
                            }

                            if (personAutoTargetData.LastBestIndex == i)
                            {
                                // increase a "consistency-score" if average magnitude is high enough, else decrease it.
                                if (avm < 0.05f)
                                {
                                    if (personAutoTargetData.LastBestIndexConsistencyScore > 0)
                                    {
                                        personAutoTargetData.LastBestIndexConsistencyScore -= 1;
                                        // SuperController.LogMessage($"C: < -- : {personAutoTargetData.LastBestIndexConsistencyScore}");
                                    }
                                }
                                else
                                {
                                    if (personAutoTargetData.LastBestIndexConsistencyScore < 125)
                                    {
                                        personAutoTargetData.LastBestIndexConsistencyScore += 1;
                                        // SuperController.LogMessage($"C: > ++ : {personAutoTargetData.LastBestIndexConsistencyScore}");
                                    }
                                }
                                score += personAutoTargetData.LastBestIndexConsistencyScore * 0.001f;
                                consistencyScoreUpdated = true;
                            }

                            if (bestMotionScore < score)
                            {
                                bestMotionScore = score;
                                bestMotionAutoTargetData = targetData;
                                bestMotionIndex = i;
                            }
                        }
                    }
                }
            }

            if (bestMotionAutoTargetData != null)
            {
                resultingAutoTargetData = bestMotionAutoTargetData;
                bestIndex = bestMotionIndex;
                motionScore = bestMotionScore;
            } 
            else 
            {
                if (resultingAutoTargetData != null)
                {
                    // closest target will do - use a score of 0 for this case
                    motionScore = 0f;
                    if (vaginalOverAnusCurrentlyPrioritized && bestIndex == 1)
                    {
                        // skip anus as long as vaginal is active, revert back to vaginal as closest
                        bestIndex = 0;
                        resultingAutoTargetData = personAutoTargetData.AutoTargets[bestIndex];
                    }
                    else if (anusOverVaginalCurrentlyPrioritized && bestIndex == 0)
                    {
                        // skip vaginal as long as anus is active, revert back to anus as closest
                        bestIndex = 1;
                        resultingAutoTargetData = personAutoTargetData.AutoTargets[bestIndex];
                    }
                }
            }

            if (bestIndex >= 0 && personAutoTargetData.LastBestIndex != bestIndex)
            {
                personAutoTargetData.LastBestIndex = bestIndex;
                personAutoTargetData.LastBestIndexConsistencyScore = 0;

                if (_lastAutoBestFemaleUid == _femaleAtom.uid)
                {
                    TCodeDevice.TargetSwappedTime = Time.time;
                    //if (ToySerialController.Plugin.DebugEnabled)
                    //    SuperController.LogMessage("TCodeDevice.TargetSwappedTime reset");
                }
                if (ToySerialController.Plugin.DebugEnabled)
                {
                    if (bestMotionAutoTargetData != null)
                        SuperController.LogMessage("Best Motion Target for " + _femaleAtom.uid + ": " + personAutoTargetData.AutoTargets[personAutoTargetData.LastBestIndex].TargetInfo + " with score (" + motionScore.ToString() + ")");
                    else
                        SuperController.LogMessage("Closest Target for " + _femaleAtom.uid + ": " + personAutoTargetData.AutoTargets[personAutoTargetData.LastBestIndex].TargetInfo);
                }
                
            } 
            else if (!consistencyScoreUpdated && personAutoTargetData.LastBestIndexConsistencyScore > 0)
            {
                personAutoTargetData.LastBestIndexConsistencyScore--;
            }

            if (resultingAutoTargetData == null)
                return false;

            //if (ToySerialController.Plugin.DebugEnabled)
            //    SuperController.LogMessage(resultingAutoTargetData.GetAverageMagnitude().ToString());

            return resultingAutoTargetData.TargetFunc();
        }

        public static Vector3 NearestPointOnLine(Vector3 linePnt, Vector3 lineDir, Vector3 pnt)
        {
            lineDir.Normalize();//this needs to be a unit vector
            var v = pnt - linePnt;
            var d = Vector3.Dot(v, lineDir);
            return linePnt + lineDir * d;
        }

        private bool UpdateVaginaTargetMale()
        {
            // Vaginal for male will be Penis
            var penisColliders = new List<string> {
                "AutoColliderGen1Hard",
                "AutoColliderGen2Hard",
                "AutoColliderGen3aHard",
                "AutoColliderGen3bHard"
            };

            var penisTransforms = _femaleAtom
                .GetComponentsInChildren<Collider>()
                .Where(c => penisColliders.Contains(c.name))
                .OrderBy(c => c.name, StringComparer.OrdinalIgnoreCase)
                .Select(c => c.transform)
                .ToList();

            if (penisTransforms.Count != 4)
                return false;

            _referenceLengthMultiplier = 1.1f;
            _referenceRadiusMultiplier = 1.1f;

            var penisLength = 0.0f;
            for (int i = 0, j = 1; j < penisTransforms.Count; i = j++)
                penisLength += Vector3.Distance(penisTransforms[i].position, penisTransforms[j].position);

            var penisOrigin = penisTransforms[0];

            var positionOffset = -penisOrigin.up * (penisTransforms[1].position - penisTransforms[0].position).magnitude * 0.15f;
            var referencePosition = penisOrigin.position + positionOffset;

            _targetUp = (referencePosition - penisOrigin.position).normalized;
            _targetRight = penisTransforms[0].forward;
            _targetForward = penisTransforms[0].right; // Vector3.Cross(_targetRight, _targetForward);

            _targetPosition = penisOrigin.position - _targetUp * (penisLength - 0.005f);

            if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            return true;
        }

        private bool UpdateVaginaTarget()
        {
            if (IncludeMales.val && _femaleAtom.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Male)
            {
                return UpdateVaginaTargetMale();
            }

            var labiaTrigger = _femaleAtom.GetRigidBodyByName("LabiaTrigger");
            var vaginaTrigger = _femaleAtom.GetRigidBodyByName("VaginaTrigger");
            var positionOffsetCollider = _femaleAtom.GetComponentByName<CapsuleCollider>("_JointB");

            if (labiaTrigger == null || vaginaTrigger == null || positionOffsetCollider == null)
                return false;

            _referenceLengthMultiplier = 1.05f;
            _referenceRadiusMultiplier = 1f;

            _targetPosition = labiaTrigger.transform.position;
            _targetUp = (vaginaTrigger.transform.position - labiaTrigger.transform.position).normalized;
            _targetRight = vaginaTrigger.transform.right;
            _targetForward = Vector3.Cross(_targetRight, _targetUp);

            var positionOffset = positionOffsetCollider.transform.position
                + positionOffsetCollider.transform.forward * positionOffsetCollider.radius
                - _targetUp * 0.0025f;
            _targetPosition += _targetUp * Vector3.Dot(positionOffset - _targetPosition, _targetUp);

            return true;
        }

        private bool UpdateFreeControllerTarget(string id)
        {
            var control = _femaleAtom.GetStorableByID(id) as FreeControllerV3;
            if (control == null)
                return false;

            var followBody = control.followWhenOffRB;
            var labiaTrigger = _femaleAtom.GetRigidBodyByName("LabiaTrigger");
            var vaginaTrigger = _femaleAtom.GetRigidBodyByName("VaginaTrigger");
            var positionOffsetCollider = _femaleAtom.GetComponentByName<CapsuleCollider>("_JointB");

            if (followBody == null || labiaTrigger == null || vaginaTrigger == null || positionOffsetCollider == null)
                return false;

            _referenceLengthMultiplier = 1f;
            _referenceRadiusMultiplier = 1f;

            var vaginaPosition = labiaTrigger.transform.position;
            var vaginaUp = (vaginaTrigger.transform.position - labiaTrigger.transform.position).normalized;
            var vaginaRight = vaginaTrigger.transform.right;
            var vaginaForward = Vector3.Cross(vaginaRight, vaginaUp);

            var positionOffset = positionOffsetCollider.transform.position
                + positionOffsetCollider.transform.forward * positionOffsetCollider.radius
                - vaginaUp * 0.0025f;
            vaginaPosition += vaginaUp * Vector3.Dot(positionOffset - vaginaPosition, vaginaUp);

            var controlToBody = control.control.position - followBody.position;
            var referenceToVagina = vaginaPosition - ReferencePosition;

            var controlToBodyRotation = Quaternion.Slerp(control.control.rotation, followBody.rotation, 0.5f).ToNormalized();
            var vaginaRotation = Quaternion.LookRotation(vaginaForward, vaginaUp);
            var rotation = Quaternion.Slerp(vaginaRotation, controlToBodyRotation, 0.125f).ToNormalized();

            _targetUp = rotation.GetUp();
            _targetRight = rotation.GetRight();
            _targetForward = rotation.GetForward();

            var maxOffsetDistance = Mathf.Min(0.05f, referenceToVagina.magnitude);
            var maxControlDistance = 0.25f;

            var controlToBodyUp = Vector3.Project(controlToBody, vaginaUp);
            var controlT = Mathf.Clamp01(controlToBodyUp.magnitude / maxControlDistance);
            var strength = controlT < 0.5 ? 2 * controlT * controlT : -1 + (4 - 2 * controlT) * controlT;
            var sign = -Mathf.Sign(Vector3.Dot(vaginaUp, controlToBodyUp));
            var offsetN = referenceToVagina.normalized;
            var offset = offsetN * sign * maxOffsetDistance * strength;

            _targetPosition = vaginaPosition - offset;
            return true;
        }

        private bool UpdateAnusTargetMale()
        {
            CapsuleCollider anusLeft = _femaleAtom.GetComponentByName<CapsuleCollider>("_JointAlMale");
            CapsuleCollider anusRight = _femaleAtom.GetComponentByName<CapsuleCollider>("_JointArMale");

            if (anusLeft == null || anusRight == null)
                return false;

            _referenceLengthMultiplier = 1.05f;
            _referenceRadiusMultiplier = 1.75f; // for better targeting from behind

            _targetUp = ((anusLeft.transform.up + anusRight.transform.up) / 2).normalized;
            _targetRight = ((anusLeft.transform.right + anusRight.transform.right) / 2).normalized;
            _targetForward = ((anusLeft.transform.forward + anusRight.transform.forward) / 2).normalized;
            _targetPosition = (anusLeft.transform.position + anusRight.transform.position) / 2;

            return true;
        }

        private bool UpdateAnusTarget()
        {
            if (IncludeMales.val && _femaleAtom.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Male)
            {
                return UpdateAnusTargetMale();
            }

            var anusLeft = _femaleAtom.GetComponentByName<CapsuleCollider>("_JointAl");
            var anusRight = _femaleAtom.GetComponentByName<CapsuleCollider>("_JointAr");

            if (anusLeft == null || anusRight == null)
                return false;

            _referenceLengthMultiplier = 1.05f;
            _referenceRadiusMultiplier = 1f;

            _targetUp = ((anusLeft.transform.up + anusRight.transform.up) / 2).normalized;
            _targetRight = ((anusLeft.transform.right + anusRight.transform.right) / 2).normalized;
            _targetForward = ((anusLeft.transform.forward + anusRight.transform.forward) / 2).normalized;
            _targetPosition = (anusLeft.transform.position + anusRight.transform.position) / 2;

            return true;
        }

        private bool UpdateMouthTarget()
        {
            var bottomLip = _femaleAtom.GetComponentByName<Transform>("lowerJawStandardColliders")?.GetComponentByName<CapsuleCollider>("_ColliderLipM");
            var topLip = _femaleAtom.GetComponentByName<Transform>("AutoCollidersTongueUpperLip")?.GetComponentByName<Transform>("AutoColliderAutoCollidersFaceCentral2Hard");
            var mouthTrigger = _femaleAtom.GetRigidBodyByName("MouthTrigger");

            if (bottomLip == null || topLip == null || mouthTrigger == null)
                return false;

            _referenceLengthMultiplier = 1f;
            _referenceRadiusMultiplier = 1f;

            var center = (topLip.transform.position + bottomLip.transform.position) / 2;
            _targetUp = (mouthTrigger.transform.position - center).normalized;
            _targetRight = mouthTrigger.transform.right;
            _targetForward = Vector3.Cross(_targetUp, _targetRight);
            _targetPosition = center - TargetUp * Vector3.Distance(center, mouthTrigger.transform.position) * 0.2f;

            DebugDraw.DrawCircle(TargetPosition, TargetUp, TargetRight, Color.gray, (topLip.transform.position - bottomLip.transform.position).magnitude / 2);

            return true;
        }

        private bool UpdateLeftHandTarget() => UpdateHandTarget("l");

        private bool UpdateRightHandTarget() => UpdateHandTarget("r");

        private bool UpdateHandTarget(string side)
        {
            var carpal = _femaleAtom.GetRigidBodyByName($"{side}Carpal2");
            var fingerBase = carpal?.GetComponentByName<CapsuleCollider>("_Collider3");
            var fingerTip = _femaleAtom.GetRigidBodyByName($"{side}Pinky3")?.GetComponentInChildren<CapsuleCollider>();

            if (carpal == null || fingerBase == null || fingerTip == null)
                return false;

            _referenceLengthMultiplier = 1.1f;
            _referenceRadiusMultiplier = 1.2f;

            float sign = 1f;
            if (side == "l")
            {
                _targetRight = -fingerBase.transform.up;
                _targetForward = -fingerBase.transform.right;
                sign = -1f;
            }
            else if (side == "r")
            {
                _targetRight = fingerBase.transform.up;
                _targetForward = fingerBase.transform.right;
            }

            var fingerBasePosition = fingerBase.transform.position - sign * fingerBase.transform.right * (fingerBase.height / 2 - fingerBase.radius) - sign * fingerBase.transform.up * fingerBase.radius;
            var fingerTipPosition = fingerTip.transform.position - fingerTip.transform.right * (fingerTip.height / 2 - fingerTip.radius) - fingerTip.transform.up * fingerTip.radius;
            _targetUp = fingerBase.transform.forward;
            _targetPosition = (fingerBasePosition + fingerTipPosition) / 2 + _targetUp * 0.04f;

            // figure out best of 4 directions to track against Reference
            if (Vector3.Dot(-_targetForward, ReferenceUp) > 0.6f)
            {
                Vector3 tmp = _targetUp;
                _targetUp = -_targetForward;
                _targetForward = tmp;
            }
            else if (Vector3.Dot(_targetForward, ReferenceUp) > 0.6f)
            {
                Vector3 tmp = _targetUp;
                _targetUp = _targetForward;
                _targetForward = tmp;
            }
            else if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            DebugDraw.DrawLine(fingerBasePosition, fingerTipPosition, Color.gray);

            return true;
        }

        private bool UpdateChestTargetMale()
        {
            CapsuleCollider left = null;
            CapsuleCollider right = null;
            CapsuleCollider chest = null;
            foreach (var comp in _femaleAtom.GetComponentsInChildren<AutoCollider>())
            {
                if (comp.name == "AutoColliderFemaleAutoColliderschest1")
                {
                    chest = comp.jointCollider as CapsuleCollider;
                }
                if (comp.name == "AutoColliderFemaleAutoColliderschest2")
                {
                    left = comp.jointCollider as CapsuleCollider;
                }
                if (comp.name == "AutoColliderFemaleAutoColliderschest2")
                {
                    right = comp.jointCollider as CapsuleCollider;
                }
            }

            if (left == null || right == null || chest == null || chest.transform == null || left.transform == null || right.transform == null)
                return false;

            _referenceLengthMultiplier = 1f;
            _referenceRadiusMultiplier = 1f;

            _targetUp = chest.transform.right;
            _targetRight = chest.transform.up;
            _targetForward = -chest.transform.forward;

            var leftPosition = left.transform.position + _targetForward * 0.02f;
            var rightPosition = right.transform.position - _targetForward * 0.02f;            
            _targetForward = (rightPosition - leftPosition).normalized;
            _targetUp = -Vector3.Cross(_targetForward, _targetRight);

            var chestPosition = chest.transform.position + _targetForward * (chest.radius + 0.2f) - _targetUp * 0.1f;

            if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            _targetPosition = Vector3.Lerp(chestPosition, (leftPosition + rightPosition) / 2, 0.3f); // + _targetUp * 0.10f;

            return true;
        }

        private bool UpdateChestTarget()
        {
            if (IncludeMales.val && _femaleAtom.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Male)
            {
                return UpdateChestTargetMale();
            }

            var left = _femaleAtom.GetComponentByName<AutoCollider>("AutoColliderFemaleAutoColliderslNipple1")?.jointCollider as CapsuleCollider;
            var right = _femaleAtom.GetComponentByName<AutoCollider>("AutoColliderFemaleAutoCollidersrNipple1")?.jointCollider as CapsuleCollider;
            var chest = _femaleAtom.GetComponentByName<AutoCollider>("AutoColliderFemaleAutoColliderschest2c")?.jointCollider as CapsuleCollider;

            if (left == null || right == null || chest == null || chest.transform == null || left.transform == null || right.transform == null)
                return false;

            _referenceLengthMultiplier = 1f;
            _referenceRadiusMultiplier = 1f;

            _targetForward = chest.transform.forward;

            var leftPosition = left.transform.position;
            var rightPosition = right.transform.position;
            var chestPosition = chest.transform.position + _targetForward * chest.radius;

            _targetRight = (rightPosition - leftPosition).normalized;
            _targetUp = Vector3.Cross(_targetForward, _targetRight);

            if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            _targetPosition = Vector3.Lerp(chestPosition, (leftPosition + rightPosition) / 2, 0.3f); // + _targetUp * 0.10f;

            return true;
        }

        private bool UpdateChestExtendedTargetMale()
        {
            CapsuleCollider left = null;
            CapsuleCollider right = null;
            CapsuleCollider chest = null;
            foreach (var comp in _femaleAtom.GetComponentsInChildren<AutoCollider>())
            {
                if (comp.name == "AutoColliderFemaleAutoColliderschest1")
                {
                    chest = comp.jointCollider as CapsuleCollider;
                }
                if (comp.name == "AutoColliderFemaleAutoColliderschest2")
                {
                    left = comp.jointCollider as CapsuleCollider;
                }
                if (comp.name == "AutoColliderFemaleAutoColliderschest2")
                {
                    right = comp.jointCollider as CapsuleCollider;
                }
            }

            if (left == null || right == null || chest == null || chest.transform == null || left.transform == null || right.transform == null)
                return false;

            _targetUp = chest.transform.right;
            _targetRight = chest.transform.up;
            _targetForward = -chest.transform.forward;

            var leftPosition = left.transform.position + _targetForward * 0.02f;
            var rightPosition = right.transform.position - _targetForward * 0.02f;
            _targetForward = (rightPosition - leftPosition).normalized;
            _targetUp = -Vector3.Cross(_targetForward, _targetRight);

            var chestPosition = chest.transform.position + _targetForward * (chest.radius + 0.2f) * 0.90f - _targetUp * 0.1f;

            float d = Vector3.Dot(_targetUp, ReferenceUp);
            if (d < 0f)
            {
                _targetUp = -_targetUp;
                d = -d;
            }

            // make the chest area bigger to allow for less precision boobs when it's aligned
            _referenceLengthMultiplier = 1f + d * 0.5f;
            _referenceRadiusMultiplier = 1f + d * 0.2f;

            _targetPosition = Vector3.Lerp(chestPosition, (leftPosition + rightPosition) / 2, 0.3f) + _targetUp * 0.05f; // chestExtended

            return true;
        }

        private bool UpdateChestExtendedTarget()
        {
            if (IncludeMales.val && _femaleAtom.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Male)
            {
                return UpdateChestExtendedTargetMale();
            }

            var left = _femaleAtom.GetComponentByName<AutoCollider>("AutoColliderFemaleAutoColliderslNipple1")?.jointCollider as CapsuleCollider;
            var right = _femaleAtom.GetComponentByName<AutoCollider>("AutoColliderFemaleAutoCollidersrNipple1")?.jointCollider as CapsuleCollider;
            var chest = _femaleAtom.GetComponentByName<AutoCollider>("AutoColliderFemaleAutoColliderschest2c")?.jointCollider as CapsuleCollider;

            if (left == null || right == null || chest == null || chest.transform == null || left.transform == null || right.transform == null)
                return false;

            _targetForward = chest.transform.forward;

            var leftPosition = left.transform.position;
            var rightPosition = right.transform.position;
            var chestPosition = chest.transform.position + _targetForward * chest.radius * 0.90f;

            _targetRight = (rightPosition - leftPosition).normalized;
            _targetUp = Vector3.Cross(_targetForward, _targetRight);

            float d = Vector3.Dot(_targetUp, ReferenceUp);
            if (d < 0f)
            {
                _targetUp = -_targetUp;
                d = -d;
            }

            // make the chest area bigger to allow for less precision boobs when it's aligned
            _referenceLengthMultiplier = 1f + d * 0.5f;
            _referenceRadiusMultiplier = 1f + d * 0.2f;

            // _targetPosition = Vector3.Lerp(chestPosition, (leftPosition + rightPosition) / 2, 0.3f) - _targetUp * 0.22f; // chestLower
            _targetPosition = Vector3.Lerp(chestPosition, (leftPosition + rightPosition) / 2, 0.3f) + _targetUp * 0.05f; // chestExtended

            return true;
        }

        private bool UpdateLeftFootTarget() => UpdateFootTarget("l");

        private bool UpdateRightFootTarget() => UpdateFootTarget("r");

        private bool UpdateFootTarget(string side)
        {
            var foot = _femaleAtom.GetRigidBodyByName($"{side}Foot");
            var footBase = foot?.GetComponentByName<CapsuleCollider>("_Collider6");

            if (foot == null || footBase == null)
                return false;

            _referenceLengthMultiplier = 1.1f;
            _referenceRadiusMultiplier = 1.3f;

            _targetRight = footBase.transform.forward;
            _targetForward = -footBase.transform.up;

            if (side == "l")
                _targetUp = footBase.transform.right;
            else if (side == "r")
                _targetUp = -footBase.transform.right;

            _targetPosition = footBase.transform.position + _targetForward * footBase.radius + _targetUp * 0.02f + _targetRight * 0.02f;

            // figure out best of 4 directions to track against Reference
            if (Vector3.Dot(-_targetRight, ReferenceUp) > 0.6f)
            {
                Vector3 tmp = _targetUp;
                _targetUp = -_targetRight;
                _targetRight = tmp;
            }
            else if (Vector3.Dot(_targetRight, ReferenceUp) > 0.6f)
            {
                Vector3 tmp = _targetUp;
                _targetUp = _targetRight;
                _targetRight = tmp;
            }
            else if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            return true;
        }

        private bool UpdateFeetTarget()
        {
            var leftFoot = _femaleAtom.GetRigidBodyByName("lFoot");
            var rightFoot = _femaleAtom.GetRigidBodyByName("rFoot");
            var leftFootBase = leftFoot?.GetComponentByName<CapsuleCollider>("_Collider6");
            var rightFootBase = rightFoot?.GetComponentByName<CapsuleCollider>("_Collider6");

            if (leftFoot == null || rightFoot == null || leftFootBase == null || rightFootBase == null)
                return false;

            var leftPosition = leftFootBase.transform.position - leftFootBase.transform.up * leftFootBase.radius;
            var rightPosition = rightFootBase.transform.position - rightFootBase.transform.up * rightFootBase.radius;

            float dia = (leftPosition - rightPosition).magnitude;
            if (dia > 0.35f) // arbitary magic value before feet are too far away for this tracker to make sense
                return false;

            _referenceLengthMultiplier = 1f;
            _referenceRadiusMultiplier = 1.2f;

            _targetPosition = (leftPosition + rightPosition) / 2;
            _targetForward = ((leftFootBase.transform.forward + rightFootBase.transform.forward) / 2).normalized;
            _targetUp = Vector3.Cross((leftPosition - rightPosition).normalized, _targetForward).normalized;
            _targetRight = Vector3.Cross(_targetUp, _targetForward).normalized;

            // flip _targetUp based on ReferenceUp for a smoother tracking
            if (Vector3.Dot(_targetUp, ReferenceUp) < 0f)
            {
                _targetUp = -_targetUp;
            }

            DebugDraw.DrawCircle(_targetPosition, Quaternion.FromToRotation(Vector3.up, _targetUp), Color.white, dia / 2);

            return true;
        }

        public const string FEMALE_NONE = "None";
        public const string FEMALE_AUTO = "*Auto*";

        private List<string> _femaleUids = new List<string>();

        private void FindFemales()
        {
            var defaultFemale = FemaleChooser.val;

            var people = Controller.GetAtoms().Where(a => a.type == "Person");
            _femaleUids = people
                .Where(a => a != null && a.GetComponentInChildren<DAZCharacterSelector>() != null && a.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Female && !HasPenis(a))
                .Select(a => a.uid)
                .ToList();

            if (IncludeMales.val)
            {
                _femaleUids.AddRange(people
                    .Where(a => a != null && a.GetComponentInChildren<DAZCharacterSelector>() != null && (a.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Male || HasPenis(a)))
                    .Select(a => a.uid));
            }

            if (defaultFemale != FEMALE_AUTO && !_femaleUids.Contains(defaultFemale))
                defaultFemale = FEMALE_AUTO; //  _femaleUids.FirstOrDefault(uid => uid == _femaleAtom?.uid) ?? _femaleUids.FirstOrDefault() ?? FEMALE_NONE;

            List<string> femaleOptions = new List<string>(_femaleUids);
            femaleOptions.Insert(0, FEMALE_AUTO);
            femaleOptions.Insert(0, FEMALE_NONE);

            FemaleChooser.choices = femaleOptions;
            FemaleChooserCallback(defaultFemale);
        }

        // Also include all Female with penis colliders (FemaleFuta plugin)
        protected HashSet<string> _penisColliders = new HashSet<string> {
                    "AutoColliderGen1Hard",
                    "AutoColliderGen2Hard",
                    "AutoColliderGen3aHard",
                    "AutoColliderGen3bHard"
                };

        protected bool HasPenis(Atom atom)
        {
            if (atom == null) return false;
            if (atom.GetComponentInChildren<DAZCharacterSelector>().gender == DAZCharacterSelector.Gender.Male) return true;

            if (atom.GetComponentsInChildren<Collider>().Where(c => c.enabled && c.gameObject.activeInHierarchy && _penisColliders.Contains(c.name)).Any())
            {
                // Deeper check to see if "AltFutaSkin" is available, it's removed when the altfute plugin is removed, but the penis colliders are not or something...
                if (atom.GetComponentsInChildren<DAZMergedSkinV2>().Where(x => {
                    // SuperController.LogMessage(x.name);
                    return x.name == "AltFutaSkin";
                    }).Any())
                {
                    return true;
                }
            }

            return false;
        }

        protected void FemaleChooserCallback(string s)
        {
            _femaleAtom = Controller.GetAtomByUid(s);
            FemaleChooser.valNoCallback = (_femaleAtom == null ? (s == FEMALE_AUTO ? FEMALE_AUTO : FEMALE_NONE) : s);
        }

        protected override void RefreshButtonCallback() 
        {
            _targetsPerPerson?.Clear();
            FindFemales();
        }
    }
}